namespace Cogwheel;

public static class COGWHEEL
{
    private static CommandsManager _commandsManager;
    private static ICogwheelConsole _console;

    public static void Initialize(CommandsManager commandsManager, ICogwheelConsole console)
    {
        _console = console;
        _commandsManager = commandsManager;
        
        _commandsManager.Initialize(_console);
        _console.Initialize(_commandsManager);
    }
    
    
    // == Public static methods for the CommandsManager == //
    public static void RunCommand(string input)
    {
        _commandsManager.RunCommand(input);
    }
    
    public static void RegisterObject(object obj)
    {
        _commandsManager.RegisterObject(obj);
    }
    
    
    // == Public static methods for the Console == //
    public static void Log(string message)
    {
        _console.Log(message);
    }
    
    public static void LogError(string message)
    {
        _console.LogError(message);
    }
    
    public static void LogWarning(string message)
    {
        _console.LogWarning(message);
    }
    
    // == Built-in commands == //
    [Command(Name = "quit", Description = "Quits the Cogwheel console.")]
    public static void QuitCogwheelConsole()
    {
        _console.Exit();
    }
    
    [Command(Name = "clear", Description = "Clears the console.")]
    public static void ClearConsole()
    {
        _console.ClearConsole();
    }
    
    [Command(Name = "help", Description = "Gets usage for given command.")]
    public static void ShowHelp(string commandName)
    {
        ICommand command = _commandsManager.GetCommandByName(commandName);
        if (command is null)
        {
            _console.LogError($"Command error: '{commandName}' is not a command.");
            return;
        }

        _console.Log($"-- Command Usage For {commandName} --\n" +
                            $"Description: {command.Description}\n" +
                            $"Usage: {_commandsManager.GetCommandUsage(command)}");
    }
    
    [Command(Name = "list", Description = "Lists all available commands.")]
    public static void ListCommands()
    {
        foreach (var command in _commandsManager.Commands)
        {
            _console.Log($"{command.Key} - {command.Value.Description}");
        }
    }
    
    [Command(Name = "what", Description = "Prints the current context.")]
    public static void ShowCurrentContext()
    {
        _console.Log($"Current context: {_commandsManager.CurrentContext.GetType()} : {_commandsManager.CurrentContextGuid}");
    }
    
    [Command(Name = "test")]
    public static void CreateTestObject(string name)
    {
        _console.Log($"Creating new TestClass object with name: {name}");
        var test = new TestClass(name);
    }
    
    [Command(Name = "find", Description = "Lists available registered objects to select from.")]
    private static void SelectContextFromRegisteredObjects(string typeName = "")
    {
        var registeredObjects = _commandsManager.RegisteredObjectInstances;
        var filteredObjects = string.IsNullOrEmpty(typeName)
            ? registeredObjects
            : registeredObjects.Where(kvp => kvp.Value.GetType().FullName == typeName).ToDictionary(kvp => kvp.Key, kvp => kvp.Value);

        if (!filteredObjects.Any())
        {
            LogError($"No registered objects found for type '{typeName}'");
            return;
        }

        Log("Available registered objects:");
        foreach (var (guid, obj) in filteredObjects)
        {
            Log($"- {obj.GetType().FullName} : {guid}");
        }
    }
    
    [Command(Name = "set", Description = "Sets the context to the specified registered object.")]
    private static void SetContextFromGuid(string guidString)
    {
        _commandsManager.SetContext(guidString);
        Log($"Set context to {_commandsManager.CurrentContext?.GetType().FullName} : {_commandsManager.CurrentContextGuid}");
    }
}