using System.Reflection;
using System.Security.Cryptography;
using System.Text;
using System.Text.RegularExpressions;
using Exception = System.Exception;

namespace Cogwheel;

public class CommandsManager
{
    protected virtual ICogwheelConsole CogwheelConsole { get; set; }
    protected virtual string CommandPattern { get; set; } = "(?<val>(\\([^\\)]+\\)))|\"(?<val>[^\"]+)\"|'(?<val>[^']+)'|(?<val>[^\\s]+)";
    
    protected virtual List<Assembly> Assemblies { get; set; } = [];
    public virtual Dictionary<string, ICommand> Commands { get; set; } = new();
    public virtual Dictionary<Type, Func<string, object>> CustomParsers { get; set; } = new();
    
    // Context related stuff
    public virtual object? CurrentContext { get; set; }
    public virtual Guid? CurrentContextGuid { get; set; }
    public virtual Dictionary<Guid, object> RegisteredObjectInstances { get; set; } = new();
    public virtual Dictionary<object, Guid> RegisteredObjectGuids { get; set; } = new();
    
    public virtual void Initialize(ICogwheelConsole console)
    {
        CogwheelConsole = console;
        RefreshCommandsList();
        CogwheelConsole.Log($"CommandsManager initialized, Commands found: {Commands.Count}");
    }
    
    
    public virtual (ICommand, List<object>)? GetCommandAndArgsFromString(string input)
    {
        var splitString = Regex.Matches(input, CommandPattern).Select(m => m.Groups["val"].Value).ToArray();
        var commandName = splitString[0];
        List<object> args = new();
        for (var splitIndex = 1; splitIndex < splitString.Length; splitIndex++)
        {
            args.Add(splitString[splitIndex]);
        }
            
        if (Commands.TryGetValue(commandName, out var command))
        {
            return (command, args);
        }

        return null;
    }

    public virtual bool RunCommand(string input, object? context = null)
    {
        if (string.IsNullOrWhiteSpace(input)) return false;
        var command = GetCommandAndArgsFromString(input);
        if (command is null)
        {
            CogwheelConsole.LogError($"Command not found: {input}");
            return false;
        }

        if (command.Value.Item1.Method.IsStatic)
        {
            return ExecuteCommand(null, command.Value);
        }

        if (context is not null && IsCommandContextValid(command.Value.Item1, context))
        {
            return ExecuteCommand(context, command.Value);
        }

        if (CurrentContext is not null && IsCommandContextValid(command.Value.Item1, CurrentContext))
        {
            return ExecuteCommand(CurrentContext, command.Value);
        }

        //CogwheelConsole.LogWarning($"Command '{command.Value.Item1.Name}' is not static and no valid context was provided, searching for first registered context of type '{command.Value.Item1.Method.DeclaringType}'");
        context = GetFirstValidRegisteredContext(command.Value.Item1.Method.DeclaringType);
        if (context is null)
        {
            CogwheelConsole.LogError($"No context of type '{command.Value.Item1.Method.DeclaringType}' found");
            return false;
        }

        if (IsCommandContextValid(command.Value.Item1, context))
        {
            return ExecuteCommand(context, command.Value);
        }

        return false;
    }

    public virtual string GetCommandUsage(ICommand command)
    {
        string paramUsage = string.Join(" ",
            command.Method.GetParameters().Select(param =>
                $"<{(param.IsDefined(typeof(ParamArrayAttribute)) ? "params " : "")}{param}>"));

        return $"{command.Name}: {paramUsage}";
    }

    public virtual bool ExecuteCommand(object obj, (ICommand command, List<object> args) command)
    {
        var parameters = command.command.Method.GetParameters();
            for (var parameterIndex = 0; parameterIndex < parameters.Length; parameterIndex++)
            {
                if (parameters[parameterIndex].IsDefined(typeof(ParamArrayAttribute), false))
                {
                    var paramList = Activator.CreateInstance(typeof(List<>).MakeGenericType(parameters[parameterIndex].ParameterType.GetElementType()!));
                    for (var argIndex = parameterIndex; argIndex < parameters.Length; argIndex++)
                    {
                        if (TryParseParameter(parameters[parameterIndex].ParameterType.GetElementType(),
                                (string)command.args[argIndex], out var val))
                        {
                            paramList.GetType().GetMethod("Add")?.Invoke(paramList, new[] { val });
                        }
                        else
                        {
                            CogwheelConsole.LogError($"Format exception: could not parse '{command.args[parameterIndex]}' as '{parameters[parameterIndex].ParameterType}'");
                            return false;
                        }
                    }

                    command.args = command.args.Take(parameterIndex).ToList();
                    command.args.Add(paramList.GetType().GetMethod("ToArray")?.Invoke(paramList, null));
                }
                else if (parameterIndex >= command.args.Count)
                {
                    if (parameters[parameterIndex].IsOptional)
                    {
                        command.args.Add(Type.Missing);
                    }
                    else
                    {
                        CogwheelConsole.LogError("Not enough args passed");
                        CogwheelConsole.Log(GetCommandUsage(command.command));
                        return false;
                    }
                }
                else
                {
                    if (TryParseParameter(parameters[parameterIndex].ParameterType, (string)command.args[parameterIndex], out var val))
                    {
                        command.args[parameterIndex] = val;
                    }
                    else
                    {
                        CogwheelConsole.LogError($"Format exception: could not parse '{command.args[parameterIndex]}' as '{parameters[parameterIndex].ParameterType}'");
                        return false;
                    }
                }
            }

            try
            {
                command.command.Method.Invoke(obj, command.args.ToArray());
                return true;
            }
            catch (Exception e)
            {
                CogwheelConsole.LogError(e.Message);
                throw;
            }
    }

    public virtual bool TryParseParameter(Type parameterType, string parameterString, out object parsedValue)
    {
        if (parameterType == typeof(string))
        {
            parsedValue = parameterString;
            return true;
        }

        if (CustomParsers.ContainsKey(parameterType))
        {
            try
            {
                parsedValue = CustomParsers[parameterType].Invoke(parameterString);
                return true;
            }
            catch (Exception e)
            {
                CogwheelConsole.LogError(e.Message);
                parsedValue = null;
                return false;
            }
        }
        else
        {
            var parseMethod = parameterType.GetMethod("Parse", new[] { typeof(string) });

            if (parseMethod is not null)
            {
                try
                {
                    parsedValue = parseMethod.Invoke(null, new object[] { parameterString });
                    return true;
                }
                catch (Exception e)
                {
                    CogwheelConsole.LogError(e.Message);
                    parsedValue = null;
                    return false;
                }
            }
        }
            
        parsedValue = null;
        return false;
    }

    public virtual ICommand? GetCommandByName(string commandName)
    {
        return Commands.GetValueOrDefault(commandName);
    }

    // Context related stuff
    
    public virtual void RegisterObject(object obj)
    {
        // Use a combination of the object's type and a hash of its properties to generate a deterministic GUID
        string seed = obj.GetType().FullName + GetObjectPropertiesHash(obj);
        using (MD5 md5 = MD5.Create())
        {
            byte[] hash = md5.ComputeHash(Encoding.UTF8.GetBytes(seed));
            Guid uniqueId = new Guid(hash.Take(16).ToArray());

            if (!RegisteredObjectInstances.ContainsKey(uniqueId))
            {
                RegisteredObjectInstances[uniqueId] = obj;
                RegisteredObjectGuids[obj] = uniqueId;
                CogwheelConsole.Log($"Registered object with ID {uniqueId}");
            }
            else
            {
                CogwheelConsole.LogWarning($"Object with ID {uniqueId} is already registered");
            }
        }
    }
    
    public virtual bool IsCommandContextValid(ICommand command, object? context = null)
    {
        if (context is not null)
        {
            return command.Method.DeclaringType == context.GetType();
        }
        return command.Method.DeclaringType == CurrentContext?.GetType();
    }

    public virtual bool SetContext(object context)
    {
        CurrentContext = context;
        if (!RegisteredObjectGuids.ContainsKey(context))
        {
            RegisterObject(context);
            return true;
        }
        CurrentContextGuid = RegisteredObjectGuids[context];
        return true;
    }

    public virtual bool SetContext(Guid guid)
    {
        if (RegisteredObjectInstances.ContainsKey(guid))
        {
            CurrentContext = RegisteredObjectInstances[guid];
            CurrentContextGuid = guid;
        }
        else
        {
            CogwheelConsole.LogError($"No object registered with ID {guid}");
            return false;
        }

        return true;
    }

    public virtual bool SetContext(string guidString)
    {
        if (Guid.TryParse(guidString, out var guid))
        {
            return SetContext(guid);
        }

        CogwheelConsole.LogError($"Could not parse '{guidString}' as a GUID");
        return false;
    }

    public virtual object? GetFirstValidRegisteredContext(Type type)
    {
        return RegisteredObjectInstances.Values.FirstOrDefault(obj => obj.GetType() == type);
    }

    public virtual Guid GetGuidFromContext(object? context = null)
    {
        if (context is not null)
        {
            return RegisteredObjectGuids[context];
        }
        
        return CurrentContextGuid ?? Guid.Empty;
    }
    
    protected virtual string GetObjectPropertiesHash(object obj)
    {
        var properties = obj.GetType().GetProperties();
        var sb = new StringBuilder();

        foreach (var prop in properties)
        {
            var value = prop.GetValue(obj)?.ToString() ?? string.Empty;
            sb.Append(value);
        }

        return sb.ToString();
    }

    protected virtual void RefreshCommandsList()
    {
        foreach (var type in Assembly.GetCallingAssembly().GetTypes())
        {
            var methods = type.GetMethods(BindingFlags.NonPublic | BindingFlags.Static | BindingFlags.Public | BindingFlags.Instance);

            foreach (var method in methods)
            {
                var attributes = method.GetCustomAttributes<CommandAttribute>();

                foreach (var attribute in attributes)
                {
                    var commandName = attribute.Name ?? method.Name;
                    var newCommand = new Command(commandName, attribute.Description, method);
                    Commands.TryAdd(newCommand.Name, newCommand);
                }
            }
        }
    }
}